diff -pruN 2.32.0+dfsg-1/.github/workflows/ci.yml 2.32.1+dfsg-1/.github/workflows/ci.yml
--- 2.32.0+dfsg-1/.github/workflows/ci.yml	2024-06-09 03:44:21.000000000 +0000
+++ 2.32.1+dfsg-1/.github/workflows/ci.yml	2024-06-27 05:04:26.000000000 +0000
@@ -139,3 +139,21 @@ jobs:
         cd build
         cmake -T clangcl ..
         cmake --build . -j $Env:NUMBER_OF_PROCESSORS
+
+  build-msys:
+    runs-on: windows-latest
+    steps:
+    - uses: actions/checkout@v3
+    - name: Setup MSYS2
+      uses: msys2/setup-msys2@v2
+      with:
+        msystem: UCRT64
+        update: true
+        pacboy: gcc-libs:p libwinpthread-git:p tbb:p zlib:p zstd:p dlfcn:p cc:p cmake:p ninja:p
+    - name: build
+      shell: msys2 {0}
+      run: |
+        mkdir build
+        cd build
+        cmake -GNinja -DMOLD_USE_MIMALLOC=OFF -DMOLD_USE_SYSTEM_TBB=ON -DMOLD_USE_MOLD=OFF ..
+        cmake --build . -j $(nproc)
diff -pruN 2.32.0+dfsg-1/CMakeLists.txt 2.32.1+dfsg-1/CMakeLists.txt
--- 2.32.0+dfsg-1/CMakeLists.txt	2024-06-09 03:44:21.000000000 +0000
+++ 2.32.1+dfsg-1/CMakeLists.txt	2024-06-27 05:04:26.000000000 +0000
@@ -44,7 +44,7 @@
 # features and behave exactly the same.
 
 cmake_minimum_required(VERSION 3.14)
-project(mold VERSION 2.32.0)
+project(mold VERSION 2.32.1)
 
 include(CMakeDependentOption)
 include(CheckSymbolExists)
@@ -58,7 +58,10 @@ endif()
 
 add_executable(mold)
 target_compile_features(mold PRIVATE cxx_std_20)
-target_link_libraries(mold PRIVATE ${CMAKE_DL_LIBS})
+
+if(MINGW)
+  target_link_libraries(mold PRIVATE dl)
+endif()
 
 if(NOT "${CMAKE_CXX_COMPILER_FRONTEND_VARIANT}" STREQUAL "MSVC")
   target_compile_options(mold PRIVATE
@@ -272,6 +275,12 @@ if(NOT APPLE AND NOT MSVC)
   target_link_options(mold PRIVATE -pthread)
 endif()
 
+# shm_open needs -lrt
+find_library(LIBRT rt)
+if(LIBRT)
+  target_link_libraries(mold PRIVATE rt)
+endif()
+
 check_symbol_exists(madvise sys/mman.h HAVE_MADVISE)
 
 # Create a .cc file containing the current git hash for `mold --version`.
@@ -322,7 +331,7 @@ list(APPEND MOLD_ELF_TEMPLATE_FILES
   elf/tls.cc
   )
 
-if(WIN32)
+if(WIN32 AND NOT MINGW)
   list(APPEND MOLD_ELF_TEMPLATE_FILES elf/lto-win32.cc)
 else()
   list(APPEND MOLD_ELF_TEMPLATE_FILES elf/lto-unix.cc)
diff -pruN 2.32.0+dfsg-1/common/common.h 2.32.1+dfsg-1/common/common.h
--- 2.32.0+dfsg-1/common/common.h	2024-06-09 03:44:21.000000000 +0000
+++ 2.32.1+dfsg-1/common/common.h	2024-06-27 05:04:26.000000000 +0000
@@ -404,9 +404,9 @@ void update_maximum(std::atomic<T> &atom
                                        std::memory_order_relaxed));
 }
 
-template <typename T, typename U>
-inline void append(std::vector<T> &vec1, std::vector<U> vec2) {
-  vec1.insert(vec1.end(), vec2.begin(), vec2.end());
+template <typename T>
+inline void append(std::vector<T> &x, const auto &y) {
+  x.insert(x.end(), y.begin(), y.end());
 }
 
 template <typename T>
@@ -963,6 +963,8 @@ class MappedFile {
 public:
   ~MappedFile() { unmap(); }
   void unmap();
+  void close_fd();
+  void reopen_fd(const std::string &path);
 
   template <typename Context>
   MappedFile *slice(Context &ctx, std::string name, u64 start, u64 size) {
diff -pruN 2.32.0+dfsg-1/common/filepath.cc 2.32.1+dfsg-1/common/filepath.cc
--- 2.32.0+dfsg-1/common/filepath.cc	2024-06-09 03:44:21.000000000 +0000
+++ 2.32.1+dfsg-1/common/filepath.cc	2024-06-27 05:04:26.000000000 +0000
@@ -36,14 +36,9 @@ std::filesystem::path to_abs_path(std::f
 
 // Returns the path of the mold executable itself
 std::string get_self_path() {
-#if __APPLE__
-    char path[8192];
-    u32 size = sizeof(path);
-    if (_NSGetExecutablePath(path, &size)) {
-      std::cerr << "_NSGetExecutablePath failed\n";
-      exit(1);
-    }
-    return path;
+#if __APPLE__ || _WIN32
+  fprintf(stderr, "mold: get_self_path is not supported");
+  exit(1);
 #elif __FreeBSD__
   // /proc may not be mounted on FreeBSD. The proper way to get the
   // current executable's path is to use sysctl(2).
diff -pruN 2.32.0+dfsg-1/common/jobs-unix.cc 2.32.1+dfsg-1/common/jobs-unix.cc
--- 2.32.0+dfsg-1/common/jobs-unix.cc	2024-06-09 03:44:21.000000000 +0000
+++ 2.32.1+dfsg-1/common/jobs-unix.cc	2024-06-27 05:04:26.000000000 +0000
@@ -8,43 +8,149 @@
 // This file implements a feature that limits the number of concurrent
 // mold processes to just 1 for each user. It is intended to be used as
 // `MOLD_JOBS=1 ninja` or `MOLD_JOBS=1 make -j$(nproc)`.
+//
+// We can't use POSIX semaphores because the counter will not be
+// decremented automatically when a process exits abnormally. That would
+// results in a deadlock. Therefore, we use lockf-based regional file
+// locking instead. Unlike POSIX semaphores, the lock will automatically
+// released on process termination.
+//
+// To wake processes that may be waiting on the lock file, we use a
+// pthread condition variable. On normal exit, mold sends notifications to
+// all waiting processes. In case of abnormal exit, we use
+// pthread_cond_timedwait so that waiters will not wait forever.
 
 #include "common.h"
 
+#include <atomic>
 #include <fcntl.h>
+#include <pthread.h>
 #include <pwd.h>
 #include <stdlib.h>
+#include <sys/mman.h>
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <unistd.h>
 
 namespace mold {
 
+static constexpr i64 MAX_JOBS = 128;
+
+struct SharedData {
+  std::atomic_bool initialized;
+  pthread_mutex_t mu;
+  pthread_cond_t cond;
+};
+
+static int num_jobs = -1;
 static int lock_fd = -1;
+static SharedData *shared_data = nullptr;
+
+static i64 get_mold_jobs() {
+  char *env = getenv("MOLD_JOBS");
+  if (!env)
+    return 0;
+
+  i64 jobs = std::stol(env);
+  if (jobs < 0)
+    return 0;
+  return std::min(jobs, MAX_JOBS);
+}
+
+static bool do_lock() {
+  for (i64 i = 0; i < num_jobs; i++) {
+    lseek(lock_fd, i, SEEK_SET);
+    if (lockf(lock_fd, F_TLOCK, 1) == 0)
+      return true;
+  }
+  return false;
+}
+
+static SharedData *get_shared_data() {
+  // Create a shared memory object and mmap it
+  std::string name = "/mold-signal-" + std::to_string(getuid());
+  i64 size = sizeof(SharedData);
+
+  int shm_fd = shm_open(name.c_str(), O_CREAT | O_RDWR, 0600);
+  if (shm_fd == -1) {
+    perror("shm_open");
+    exit(1);
+  }
+
+  if (ftruncate(shm_fd, size) == -1) {
+    perror("ftruncate");
+    exit(1);
+  }
+
+  SharedData *data = (SharedData *)mmap(0, size, PROT_READ | PROT_WRITE,
+                                        MAP_SHARED, shm_fd, 0);
+  close(shm_fd);
+
+  if (data->initialized.exchange(true) == false) {
+    pthread_mutexattr_t mu_attr;
+    pthread_mutexattr_init(&mu_attr);
+    pthread_mutexattr_setpshared(&mu_attr, PTHREAD_PROCESS_SHARED);
+
+#ifndef __APPLE__
+    pthread_mutexattr_setrobust(&mu_attr, PTHREAD_MUTEX_ROBUST);
+#endif
+
+    pthread_mutex_init(&data->mu, &mu_attr);
+
+    pthread_condattr_t cond_attr;
+    pthread_condattr_init(&cond_attr);
+    pthread_condattr_setpshared(&cond_attr, PTHREAD_PROCESS_SHARED);
+    pthread_cond_init(&data->cond, &cond_attr);
+  }
+  return data;
+}
 
 void acquire_global_lock() {
-  char *jobs = getenv("MOLD_JOBS");
-  if (!jobs || jobs != "1"s)
+  num_jobs = get_mold_jobs();
+  if (num_jobs == 0)
     return;
 
+  shared_data = get_shared_data();
+
   std::string path;
   if (char *dir = getenv("XDG_RUNTIME_DIR"))
-    path = dir + "/mold-lock"s;
+    path = dir + "/mold.lock"s;
   else
-    path = "/tmp/mold-lock-"s + getpwuid(getuid())->pw_name;
+    path = "/tmp/mold-" + std::to_string(getuid()) + ".lock";
 
-  int fd = open(path.c_str(), O_WRONLY | O_CREAT | O_CLOEXEC, 0600);
-  if (fd == -1)
+  lock_fd = open(path.c_str(), O_WRONLY | O_CREAT | O_CLOEXEC, 0600);
+  if (lock_fd == -1 || do_lock())
     return;
 
-  if (lockf(fd, F_LOCK, 0) == -1)
-    return;
-  lock_fd = fd;
+  pthread_mutex_t *mu = &shared_data->mu;
+  pthread_cond_t *cond = &shared_data->cond;
+  int r = pthread_mutex_lock(mu);
+
+#ifndef __APPLE__
+  // If the previous process got killed while holding the mutex, the
+  // mutex has became inconsistent. We need to fix it in that case.
+  if (r == EOWNERDEAD)
+    pthread_mutex_consistent(mu);
+#endif
+
+  for (;;) {
+    struct timespec ts;
+    clock_gettime(CLOCK_REALTIME, &ts);
+    ts.tv_sec += 1;
+
+    int r = pthread_cond_timedwait(cond, mu, &ts);
+    if (do_lock() || r != ETIMEDOUT)
+      break;
+  }
+
+  pthread_mutex_unlock(mu);
 }
 
 void release_global_lock() {
-  if (lock_fd != -1)
-    close(lock_fd);
+  if (lock_fd == -1)
+    return;
+  close(lock_fd);
+  pthread_cond_broadcast(&shared_data->cond);
 }
 
 } // namespace mold
diff -pruN 2.32.0+dfsg-1/common/mapped-file-unix.cc 2.32.1+dfsg-1/common/mapped-file-unix.cc
--- 2.32.0+dfsg-1/common/mapped-file-unix.cc	2024-06-09 03:44:21.000000000 +0000
+++ 2.32.1+dfsg-1/common/mapped-file-unix.cc	2024-06-27 05:04:26.000000000 +0000
@@ -32,9 +32,20 @@ MappedFile *open_file_impl(const std::st
 void MappedFile::unmap() {
   if (size == 0 || parent || !data)
     return;
-
   munmap(data, size);
   data = nullptr;
 }
 
+void MappedFile::close_fd() {
+  if (fd == -1)
+    return;
+  close(fd);
+  fd = -1;
+}
+
+void MappedFile::reopen_fd(const std::string &path) {
+  if (fd == -1)
+    fd = open(path.c_str(), O_RDONLY);
+}
+
 } // namespace mold
diff -pruN 2.32.0+dfsg-1/common/mapped-file-win32.cc 2.32.1+dfsg-1/common/mapped-file-win32.cc
--- 2.32.0+dfsg-1/common/mapped-file-win32.cc	2024-06-09 03:44:21.000000000 +0000
+++ 2.32.1+dfsg-1/common/mapped-file-win32.cc	2024-06-27 05:04:26.000000000 +0000
@@ -64,4 +64,18 @@ void MappedFile::unmap() {
   data = nullptr;
 }
 
+void MappedFile::close_fd() {
+  if (fd == INVALID_HANDLE_VALUE)
+    return;
+  CloseHandle(fd);
+  fd = INVALID_HANDLE_VALUE;
+}
+
+void MappedFile::reopen_fd(const std::string &path) {
+  if (fd == INVALID_HANDLE_VALUE)
+    fd = CreateFileA(path.c_str(), GENERIC_READ,
+                     FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
+                     nullptr, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, nullptr);
+}
+
 } // namespace mold
diff -pruN 2.32.0+dfsg-1/common/signal-win32.cc 2.32.1+dfsg-1/common/signal-win32.cc
--- 2.32.0+dfsg-1/common/signal-win32.cc	2024-06-09 03:44:21.000000000 +0000
+++ 2.32.1+dfsg-1/common/signal-win32.cc	2024-06-27 05:04:26.000000000 +0000
@@ -38,6 +38,10 @@ static LONG WINAPI vectored_handler(_EXC
     static const char msg[] =
       "mold: failed to write to an output file. Disk full?\n";
     (void)!_write(_fileno(stderr), msg, sizeof(msg) - 1);
+  } else if (rec->ExceptionCode == EXCEPTION_STACK_OVERFLOW) {
+    static const char msg[] =
+      "mold: stack overflow\n";
+    (void)!_write(_fileno(stderr), msg, sizeof(msg) - 1);
   }
 
   cleanup();
diff -pruN 2.32.0+dfsg-1/debian/changelog 2.32.1+dfsg-1/debian/changelog
--- 2.32.0+dfsg-1/debian/changelog	2024-06-13 21:13:30.000000000 +0000
+++ 2.32.1+dfsg-1/debian/changelog	2024-06-27 06:48:59.000000000 +0000
@@ -1,3 +1,9 @@
+mold (2.32.1+dfsg-1) unstable; urgency=medium
+
+  * New upstream release
+
+ -- Sylvestre Ledru <sylvestre@debian.org>  Thu, 27 Jun 2024 08:48:59 +0200
+
 mold (2.32.0+dfsg-1) unstable; urgency=medium
 
   * New upstream release
diff -pruN 2.32.0+dfsg-1/docs/mold.md 2.32.1+dfsg-1/docs/mold.md
--- 2.32.0+dfsg-1/docs/mold.md	2024-06-09 03:44:21.000000000 +0000
+++ 2.32.1+dfsg-1/docs/mold.md	2024-06-27 05:04:26.000000000 +0000
@@ -832,8 +832,6 @@ but as `-o magic`.
   consider setting this environment variable to `1` to see if it addresses the
   OOM issue.
 
-  Currently, any value other than `1` is silently ignored.
-
 * `MOLD_DEBUG`:
   If this variable is set to a non-empty string, `mold` embeds its
   command-line options in the output file's `.comment` section.
diff -pruN 2.32.0+dfsg-1/elf/arch-riscv.cc 2.32.1+dfsg-1/elf/arch-riscv.cc
--- 2.32.0+dfsg-1/elf/arch-riscv.cc	2024-06-09 03:44:21.000000000 +0000
+++ 2.32.1+dfsg-1/elf/arch-riscv.cc	2024-06-27 05:04:26.000000000 +0000
@@ -821,7 +821,7 @@ void InputSection<E>::scan_relocations(C
       break;
     case R_RISCV_64:
       if constexpr (!E::is_64)
-        Fatal(ctx) << *this << ": R_RISCV_64 cannot be used on RV32";
+        Error(ctx) << *this << ": R_RISCV_64 cannot be used on RV32";
       scan_dyn_absrel(ctx, sym, rel);
       break;
     case R_RISCV_CALL:
@@ -854,7 +854,7 @@ void InputSection<E>::scan_relocations(C
       break;
     case R_RISCV_GPREL_HI20:
       if (ctx.arg.shared)
-        Fatal(ctx) << *this << ": R_RISCV_GPREL_HI20 may not be used with -shared";
+        Error(ctx) << *this << ": R_RISCV_GPREL_HI20 may not be used with -shared";
       break;
     case R_RISCV_BRANCH:
     case R_RISCV_JAL:
@@ -1217,7 +1217,7 @@ struct Extn {
 //
 // This function returns true if the first extension name should precede
 // the second one as per the rule.
-static bool extn_name_less(const Extn &e1, const Extn &e2) {
+static bool extn_name_less(std::string_view x, std::string_view y) {
   auto get_single_letter_rank = [](char c) -> i64 {
     std::string_view exts = "iemafdqlcbkjtpvnh";
     size_t pos = exts.find_first_of(c);
@@ -1239,57 +1239,26 @@ static bool extn_name_less(const Extn &e
     }
   };
 
-  return std::tuple{get_rank(e1.name), e1.name} <
-         std::tuple{get_rank(e2.name), e2.name};
-}
-
-static bool extn_version_less(const Extn &e1, const Extn &e2) {
-  return std::tuple{e1.major, e1.minor} <
-         std::tuple{e2.major, e2.minor};
-}
-
-static std::optional<Extn> read_extn_string(std::string_view &str) {
-  auto flags = std::regex_constants::optimize | std::regex_constants::ECMAScript;
-  static std::regex re(R"(^([a-z]+)(\d+)p(\d+))", flags);
-
-  std::cmatch m;
-  if (std::regex_search(str.data(), str.data() + str.size(), m, re)) {
-    str = str.substr(m.length());
-    return Extn{m[1], (i64)std::stoul(m[2]), (i64)std::stoul(m[3])};
-  }
-  return {};
+  return std::tuple{get_rank(x), x} < std::tuple{get_rank(y), y};
 }
 
 static std::vector<Extn> parse_arch_string(std::string_view str) {
-  if (str.size() < 5)
-    return {};
-
-  // Parse the base part
-  std::string_view base = str.substr(0, 5);
-  if (base != "rv32i" && base != "rv32e" && base != "rv64i" && base != "rv64e")
-    return {};
-  str = str.substr(4);
-
-  std::optional<Extn> extn = read_extn_string(str);
-  if (!extn)
-    return {};
+  auto flags = std::regex_constants::optimize | std::regex_constants::ECMAScript;
+  static std::regex re(R"(^([a-z]|[a-z][a-z0-9]*[a-z])(\d+)p(\d+)(_|$))", flags);
 
   std::vector<Extn> vec;
-  extn->name = base;
-  vec.push_back(*extn);
 
-  // Parse extensions
-  while (!str.empty()) {
-    if (str[0] != '_')
+  for (;;) {
+    std::cmatch m;
+    if (!std::regex_search(str.data(), str.data() + str.size(), m, re))
       return {};
-    str = str.substr(1);
 
-    std::optional<Extn> extn = read_extn_string(str);
-    if (!extn)
-      return {};
-    vec.push_back(*extn);
+    vec.push_back(Extn{m[1], (i64)std::stoul(m[2]), (i64)std::stoul(m[3])});
+    if (m[4].length() == 0)
+      return vec;
+
+    str = str.substr(m.length());
   }
-  return vec;
 }
 
 static std::vector<Extn> merge_extensions(std::span<Extn> x, std::span<Extn> y) {
@@ -1302,10 +1271,13 @@ static std::vector<Extn> merge_extension
   // Merge ISA extension strings
   while (!x.empty() && !y.empty()) {
     if (x[0].name == y[0].name) {
-      vec.push_back(extn_version_less(x[0], y[0]) ? y[0] : x[0]);
+      if (std::tuple{x[0].major, x[0].minor} < std::tuple{y[0].major, y[0].minor})
+        vec.push_back(y[0]);
+      else
+        vec.push_back(x[0]);
       x = x.subspan(1);
       y = y.subspan(1);
-    } else if (extn_name_less(x[0], y[0])) {
+    } else if (extn_name_less(x[0].name, y[0].name)) {
       vec.push_back(x[0]);
       x = x.subspan(1);
     } else {
@@ -1314,18 +1286,19 @@ static std::vector<Extn> merge_extension
     }
   }
 
-  vec.insert(vec.end(), x.begin(), x.end());
-  vec.insert(vec.end(), y.begin(), y.end());
+  append(vec, x);
+  append(vec, y);
   return vec;
 }
 
-static std::string to_string(std::span<Extn> v) {
-  std::string str = v[0].name + std::to_string(v[0].major) + "p" +
-                    std::to_string(v[0].minor);
+static std::string to_string(const Extn &e) {
+  return e.name + std::to_string(e.major) + "p" + std::to_string(e.minor);
+}
 
+static std::string to_string(std::span<Extn> v) {
+  std::string str = to_string(v[0]);
   for (i64 i = 1; i < v.size(); i++)
-    str += "_" + v[i].name + std::to_string(v[i].major) + "p" +
-           std::to_string(v[i].minor);
+    str += "_" + to_string(v[i]);
   return str;
 }
 
@@ -1408,7 +1381,7 @@ void RiscvAttributesSection<E>::update_s
 
 template <>
 void RiscvAttributesSection<E>::copy_buf(Context<E> &ctx) {
-  memcpy(ctx.buf + this->shdr.sh_offset, contents.data(), contents.size());
+  write_vector(ctx.buf + this->shdr.sh_offset, contents);
 }
 
 } // namespace mold::elf
diff -pruN 2.32.0+dfsg-1/elf/cmdline.cc 2.32.1+dfsg-1/elf/cmdline.cc
--- 2.32.0+dfsg-1/elf/cmdline.cc	2024-06-09 03:44:21.000000000 +0000
+++ 2.32.1+dfsg-1/elf/cmdline.cc	2024-06-27 05:04:26.000000000 +0000
@@ -8,11 +8,11 @@
 #include <tbb/global_control.h>
 #include <unordered_set>
 
-#ifdef _WIN32
+#if __has_include(<unistd.h>)
+# include <unistd.h>
+#else
 # define isatty _isatty
 # define STDERR_FILENO (_fileno(stderr))
-#else
-# include <unistd.h>
 #endif
 
 namespace mold::elf {
@@ -542,17 +542,21 @@ std::vector<std::string> parse_nonpositi
   if constexpr (is_riscv<E>)
     ctx.arg.discard_locals = true;
 
-  // It looks like the SPARC's dynamic linker takes both RELA's r_addend
-  // and the value at the relocated place. So we don't want to write
-  // values to relocated places.
-  if constexpr (is_sparc<E>)
+  // We generally don't need to write addends to relocated places if the
+  // relocation type is RELA because RELA records contain addends.
+  // However, there are too much code that wrongly assumes that addends
+  // are written to both RELA records and relocated places, so we write
+  // addends to relocated places by default. There are a few exceptions:
+  //
+  // - It looks like the SPARC's dynamic linker takes both RELA's r_addend
+  //   and the value at the relocated place. So we don't want to write
+  //   values to relocated places.
+  //
+  // - Static PIE binaries crash on startup in some RISC-V environment if
+  //   we write addends to relocated places.
+  if constexpr (is_sparc<E> || is_riscv<E>)
     ctx.arg.apply_dynamic_relocs = false;
 
-  // For some reason, SH4 always stores relocation addends to
-  // relocated places even though its RELA.
-  if constexpr (is_sh4<E>)
-    ctx.arg.apply_dynamic_relocs = true;
-
   auto read_arg = [&](std::string name) {
     for (const std::string &opt : add_dashes(name)) {
       if (args[0] == opt) {
@@ -1360,6 +1364,9 @@ std::vector<std::string> parse_nonpositi
       Fatal(ctx) << "-auxiliary may not be used without -shared";
   }
 
+  // Even though SH4 is RELA, addends in its relocation records are always
+  // zero, and actual addends are written to relocated places. So we need
+  // to handle it as an exception.
   if constexpr (!E::is_rela || is_sh4<E>)
     if (!ctx.arg.apply_dynamic_relocs)
       Fatal(ctx) << "--no-apply-dynamic-relocs may not be used on "
diff -pruN 2.32.0+dfsg-1/elf/input-files.cc 2.32.1+dfsg-1/elf/input-files.cc
--- 2.32.0+dfsg-1/elf/input-files.cc	2024-06-09 03:44:21.000000000 +0000
+++ 2.32.1+dfsg-1/elf/input-files.cc	2024-06-27 05:04:26.000000000 +0000
@@ -1072,14 +1072,16 @@ ObjectFile<E>::mark_live_objects(Context
     if (sym.is_traced)
       print_trace_symbol(ctx, *this, esym, sym);
 
-    if (sym.file && !esym.is_weak() &&
-        (esym.is_undef() || (esym.is_common() && !sym.esym().is_common())) &&
-        !sym.file->is_alive.test_and_set()) {
-      feeder(sym.file);
+    if (sym.file) {
+      bool undef_ref = esym.is_undef() && (!esym.is_weak() || sym.file->is_dso);
+      bool common_ref = esym.is_common() && !sym.esym().is_common();
 
-      if (sym.is_traced)
-        Out(ctx) << "trace-symbol: " << *this << " keeps " << *sym.file
-                 << " for " << sym;
+      if ((undef_ref || common_ref) && !sym.file->is_alive.test_and_set()) {
+        feeder(sym.file);
+        if (sym.is_traced)
+          Out(ctx) << "trace-symbol: " << *this << " keeps " << *sym.file
+                   << " for " << sym;
+      }
     }
   }
 }
diff -pruN 2.32.0+dfsg-1/elf/lto-unix.cc 2.32.1+dfsg-1/elf/lto-unix.cc
--- 2.32.0+dfsg-1/elf/lto-unix.cc	2024-06-09 03:44:21.000000000 +0000
+++ 2.32.1+dfsg-1/elf/lto-unix.cc	2024-06-27 05:04:26.000000000 +0000
@@ -200,10 +200,7 @@ static PluginStatus release_input_file(c
   LOG << "release_input_file\n";
 
   ObjectFile<E> &file = *(ObjectFile<E> *)handle;
-  if (file.mf->fd != -1) {
-    close(file.mf->fd);
-    file.mf->fd = -1;
-  }
+  file.mf->close_fd();
   return LDPS_OK;
 }
 
@@ -449,7 +446,7 @@ static void load_lto_plugin(Context<E> &
     phase = 1;
     gctx<E> = &ctx;
 
-    void *handle = dlopen(ctx.arg.plugin.c_str(), RTLD_NOW | RTLD_GLOBAL);
+    void *handle = dlopen(ctx.arg.plugin.c_str(), RTLD_NOW | RTLD_LOCAL);
     if (!handle)
       Fatal(ctx) << "could not open plugin file: " << dlerror();
 
@@ -570,7 +567,11 @@ static ElfSym<E> to_elf_sym(PluginSymbol
 // Returns false if it's GCC.
 template <typename E>
 static bool is_llvm(Context<E> &ctx) {
+#ifdef __MINGW32__
+  return ctx.arg.plugin.ends_with("LLVMgold.dll");
+#else
   return ctx.arg.plugin.ends_with("LLVMgold.so");
+#endif
 }
 
 // Returns true if a given linker plugin supports the get_symbols_v3 API.
@@ -590,11 +591,11 @@ create_plugin_input_file(Context<E> &ctx
   file.offset = mf->get_offset();
   file.filesize = mf->size;
 
-  if (mf2->fd == -1)
-    mf2->fd = open(file.name, O_RDONLY);
+  mf2->reopen_fd(file.name);
+
   file.fd = mf2->fd;
 
-  if (file.fd == -1)
+  if (!file.fd)
     Fatal(ctx) << "cannot open " << file.name << ": " << errno_string();
   return file;
 }
@@ -645,8 +646,7 @@ ObjectFile<E> *read_lto_object(Context<E
   // open files" issue, we close fd only for GCC. This is ugly, though.
   if (!is_llvm(ctx)) {
     MappedFile *mf2 = mf->parent ? mf->parent : mf;
-    close(mf2->fd);
-    mf2->fd = -1;
+    mf2->close_fd();
   }
 
   // Create a symbol strtab
diff -pruN 2.32.0+dfsg-1/elf/lto.h 2.32.1+dfsg-1/elf/lto.h
--- 2.32.0+dfsg-1/elf/lto.h	2024-06-09 03:44:21.000000000 +0000
+++ 2.32.1+dfsg-1/elf/lto.h	2024-06-27 05:04:26.000000000 +0000
@@ -73,7 +73,11 @@ enum PluginOutputFileType {
 
 struct PluginInputFile {
   const char *name;
-  i32 fd;
+#if __MINGW32__
+  HANDLE fd;
+#else
+  int fd;
+#endif
   u64 offset;
   u64 filesize;
   void *handle;
diff -pruN 2.32.0+dfsg-1/test/elf/common.inc 2.32.1+dfsg-1/test/elf/common.inc
--- 2.32.0+dfsg-1/test/elf/common.inc	2024-06-09 03:44:21.000000000 +0000
+++ 2.32.1+dfsg-1/test/elf/common.inc	2024-06-27 05:04:26.000000000 +0000
@@ -69,6 +69,10 @@ test_cflags() {
   echo 'int main() {}' | $CC "$@" -o /dev/null -xc - >& /dev/null
 }
 
+test_cxxflags() {
+  echo 'int main() {}' | $CXX "$@" -o /dev/null -xc++ - >& /dev/null
+}
+
 is_musl() {
   ldd --version 2>&1 | grep -q musl
 }
diff -pruN 2.32.0+dfsg-1/test/elf/exception.sh 2.32.1+dfsg-1/test/elf/exception.sh
--- 2.32.0+dfsg-1/test/elf/exception.sh	2024-06-09 03:44:21.000000000 +0000
+++ 2.32.1+dfsg-1/test/elf/exception.sh	2024-06-27 05:04:26.000000000 +0000
@@ -5,7 +5,7 @@
 [ $MACHINE = sh4 ] && skip
 
 static=
-test_cflags -static && static=-static
+test_cxxflags -static && static=-static
 
 # I don't know why, but we need -pthread on m68k
 static="$static -pthread"
diff -pruN 2.32.0+dfsg-1/test/elf/mold-jobs.sh 2.32.1+dfsg-1/test/elf/mold-jobs.sh
--- 2.32.0+dfsg-1/test/elf/mold-jobs.sh	1970-01-01 00:00:00.000000000 +0000
+++ 2.32.1+dfsg-1/test/elf/mold-jobs.sh	2024-06-27 05:04:26.000000000 +0000
@@ -0,0 +1,23 @@
+#!/bin/bash
+. $(dirname $0)/common.inc
+
+# Looks like lockf doesn't work correctly on qemu-riscv64
+[ $MACHINE = riscv64 -o $MACHINE = riscv32 ] && skip
+
+cat <<EOF | $CC -o $t/a.o -c -xc - -fno-PIE
+#include <stdio.h>
+int main() {
+  printf("Hello world\n");
+}
+EOF
+
+for i in `seq 1 20`; do
+  rm -f $t/exe$i
+  ( MOLD_JOBS=2 $CC -B. -o $t/exe$i $t/a.o -no-pie; echo $i) &
+done
+
+wait
+
+for i in `seq 1 20`; do
+  $QEMU $t/exe$i | grep -q 'Hello world'
+done
diff -pruN 2.32.0+dfsg-1/test/elf/weak-export-dso2.sh 2.32.1+dfsg-1/test/elf/weak-export-dso2.sh
--- 2.32.0+dfsg-1/test/elf/weak-export-dso2.sh	2024-06-09 03:44:21.000000000 +0000
+++ 2.32.1+dfsg-1/test/elf/weak-export-dso2.sh	2024-06-27 05:04:26.000000000 +0000
@@ -18,4 +18,4 @@ int main() {
 EOF
 
 $CC -B. -o $t/d.so $t/c.o $t/b.so -shared
-readelf -W --dyn-syms $t/d.so | grep -q 'WEAK   DEFAULT  UND foo'
+readelf -W --dyn-syms $t/d.so | grep -q 'WEAK   DEFAULT .* UND foo'
diff -pruN 2.32.0+dfsg-1/test/elf/weak-undef5.sh 2.32.1+dfsg-1/test/elf/weak-undef5.sh
--- 2.32.0+dfsg-1/test/elf/weak-undef5.sh	1970-01-01 00:00:00.000000000 +0000
+++ 2.32.1+dfsg-1/test/elf/weak-undef5.sh	2024-06-27 05:04:26.000000000 +0000
@@ -0,0 +1,21 @@
+#!/bin/bash
+. $(dirname $0)/common.inc
+
+cat <<EOF | $CC -c -o $t/a.o -fPIC -xc -
+#include <stdio.h>
+__attribute__((weak)) int foo();
+int main() {
+  printf("%d\n", foo ? foo() : -1);
+}
+EOF
+
+cat <<EOF | $CC -c -o $t/b.o -fPIC -xc -
+#include <stdio.h>
+int foo() { return 2; }
+EOF
+
+$CC -B. -o $t/libfoobar.so $t/b.o -shared
+$CC -B. -o $t/exe $t/a.o -Wl,--as-needed -L$t -lfoobar -Wl,-rpath,$t
+
+readelf --dynamic $t/exe | grep -q 'NEEDED.*libfoobar'
+$QEMU $t/exe | grep -q '^2$'
diff -pruN 2.32.0+dfsg-1/test/elf/x86_64_exception-mcmodel-large.sh 2.32.1+dfsg-1/test/elf/x86_64_exception-mcmodel-large.sh
--- 2.32.0+dfsg-1/test/elf/x86_64_exception-mcmodel-large.sh	2024-06-09 03:44:21.000000000 +0000
+++ 2.32.1+dfsg-1/test/elf/x86_64_exception-mcmodel-large.sh	2024-06-27 05:04:26.000000000 +0000
@@ -15,7 +15,7 @@ EOF
 $CXX -B. -o $t/exe $t/a.o -mcmodel=large
 $QEMU $t/exe
 
-if echo 'int main() {}' | $CC -o /dev/null -xc - -static >& /dev/null; then
+if test_cxxflags -static; then
   $CXX -B. -o $t/exe $t/a.o -static -mcmodel=large
   $QEMU $t/exe
 fi
diff -pruN 2.32.0+dfsg-1/third-party/mimalloc/src/prim/unix/prim.c 2.32.1+dfsg-1/third-party/mimalloc/src/prim/unix/prim.c
--- 2.32.0+dfsg-1/third-party/mimalloc/src/prim/unix/prim.c	2024-06-09 03:44:21.000000000 +0000
+++ 2.32.1+dfsg-1/third-party/mimalloc/src/prim/unix/prim.c	2024-06-27 05:04:26.000000000 +0000
@@ -28,6 +28,7 @@ terms of the MIT license. A copy of the
 #include <sys/mman.h>  // mmap
 #include <unistd.h>    // sysconf
 #include <fcntl.h>     // open, close, read, access
+#include <stdlib.h>
   
 #if defined(__linux__)
   #include <features.h>
